/*2021-03-01T17:48:00.467Z*/

/**********************************************************************************************************************
 * esp_alg.c
 *
 * generated by : embedded software producer/EfmiProductionCodeGen 1.0.0
 *
 * with types :
 *   Boolean   = unsigned int
 *   Integer   = signed int
 *   Real      = float
 *
 * with calling convention :
 *   Arguments = global var (efmiData)
 *   Ret. type = Boolean
 *
 * with names :
 *   Init Fcn  = efmiInitialize
 *   Step Fcn  = efmiDoStep
 *   Recal Fcn = efmiReInitialize
 *
 * with extra comments for :
 *   Code     = true
 *   Location = true
 *
 * with optimizations for :
 *   Indexing = true
 *   Parenth. = true (Misra: true)
 *   Id. len  = -1
 *   Tab size = 4
 *   Line len = 120
 **********************************************************************************************************************/

#include "esp_alg.h"

#ifdef __cplusplus
extern "C" {
#endif

static void pivot(
    const Integer n,
    Real* const A,
    const Integer col_ind,
    Integer* const p);
static Integer LinFactorize(
    const Integer n,
    Real* const A,
    const Integer nw,
    Real* const work,
    Integer* const p);
static Integer LinSolve(
    const Integer n,
    const Real* const D,
    const Integer* const p,
    Real* const b,
    const Integer nw,
    Real* const work);

/* Change the definition here if you don't accept float comparison to zero */
static inline Boolean isZero(Real x)
{
    return absolute(x) == real(0);
}

/*
 * Gets proper NaN from strtoX() api (C99 + POSIX2001)
 */
Real ESP_nan(void)
{
    static Boolean set = false;
    static Real nan;
    if(true != set)
    {
        nan = FLT_SEL(strtof,strtod,strtold)("NAN",NULL);
        /* if C99 api fails (eg. with tcc), the rever to the old & ugly 0/0 way of things,
         * hoping for the compiler not complaining */
        if(!isnan(nan))
        {
            nan -= nan;
            nan /= nan;
        }
        set = isnan(nan) ? true : false; /* should be true */
    }
    return nan;
}

/*
    Find entry along column col_ind of A (from the row with index col_ind and below) with largest
    modulus. Then swaps the row with index col_ind with the row containing the largest entry. The
    reordering is stored in the pivot vector p.

    -> Integer  n       : Dimensions of A and p; n > 0 required
    -> Real     A[n, n] : A-matrix stored in one vector in row-major order
    <- Real     A[n, n] : Pivoted A-matrix
    -> Integer  col_ind : Index of column to search in
    -> Integer  p       : Permutation vector containing the order of the rows of A
    <- Integer  p       : Permutation vector after pivoting
*/
static void pivot(
    const Integer n,
    Real* const A,
    const Integer col_ind,
    Integer* const p)
{
    Integer max_ind = col_ind, j;
    Real max_val = absolute(A[col_ind*n + col_ind]);

    /* Find entry with largest modulus in column col_ind (row col_ind and below). */
    for (j = col_ind + 1; j != n; ++j)
    {
        Real val = absolute(A[j*n + col_ind]);
        if (val > max_val)
        {
            max_val = val;
            max_ind = j;
        }
    }

    /* If largest entry found below row col_ind, update p and swap the rows. */
    if (max_ind != col_ind)
    {
        const Integer temp_p = p[max_ind];

        p[max_ind] = p[col_ind];
        p[col_ind] = temp_p;
        for (j = 0; j != n; ++j)
        {
            const Real temp_A = A[max_ind*n + j];
            A[max_ind*n + j] = A[col_ind*n + j];
            A[col_ind*n + j] = temp_A;
        }
    }
}

/*
    The function performs an LU factorization of matrix A using partial pivoting. The matrices L
    and U are returned in A. The upper-triangular U in the upper-triangular part of A
    (including the main diagonal). The lower-triangular L in the lower- triangular part of A. L
    is 1 on the main diagonal, however these values are not stored. Partial pivoting is used to
    give below-diagonal elements in L of magnitude less than 1. The order of the rows after
    pivoting is returned in the vector p.

        Algorithm based on:
        Trefethen, L.N., Bau, D. (1997)
        Numerical Linear Algebra
        Lecture 21

    -> Integer  n       : Dimensions of A and p; n > 0 required
    -> Real     A[n, n] : A-matrix stored in one vector in row-major order
    <- Real     A[n, n] : LU-decomposition of A
    -> Integer  nw      : Length of work array (not needed here)
    -> Real     work    : Real work array of length nw (not needed here)
    <- Integer  p[n]    : Pivot vector
    <- Return           : =  0: No error: LU decomposition successfully computed.
                          = -1: Error: A is singular
*/
static Integer LinFactorize(
    const Integer n,
    Real* const A,
    const Integer nw,
    Real* const work,
    Integer* const p)
{
    Integer k, j, i;

    /* Initialize the pivot vector p. */
    for (k = 0; k != n; ++k)
    {
        p[k] = k;
    }

    /* LU algorithm with partial pivoting. */
    for (k = 0; k != n - 1; ++k)
    {
        const Integer kn = k * n;

        /*
            Find entry along column k (row k and below) with largest modulus and swap row k with
            the row containing the largest entry.
        */
        pivot(n, A, k, p);

        /* The pivot element is zero iff the matrix A is singular. */
        if (isZero(A[kn + k]))
        {
            return -1;
        }

        /* Perform elimination on all rows below the current (row k). */
        for (j = k + 1; j != n; ++j)
        {
            const Integer jn = j * n;

            A[jn + k] /= A[kn + k];
            for (i = k + 1; i != n; ++i)
            {
                A[jn + i] -= A[jn + k] * A[kn + i];
            }
        }
    }

    /* If the last diagonal element is zero after decomposition the matrix is singular. */
    if (isZero(A[(n*n) - 1]))
    {
        return -1;
    }

    /* If non of the singularity checks above are valid the matrix is non-singular. */
    return 0;
}

/*
    The matrix D is assumed to contain the factors L and U and the row order information is assumed
    to be stored in the vector p. The right-hand-side is given in b and the solution is returned in
    the same vector.

    -> Integer  n       : Dimensions of D and pivots; n > 0 required
    -> Real     D[n, n] : LU-decomposition of A stored in one vector in row-major order
    -> Integer  p[n]    : Pivot vector
    -> Real     b[n]    : b-vector
    <- Real     b[n]    : solution vector of linear system
    -> Integer  nw      : Length of work array
    -> Real     work[n] : Real work array of length nw
    <- Return           : =  0: No error: Solution of linear system successfully computed.
                          = -1: Error.
*/
static Integer LinSolve(
    const Integer n,
    const Real* const D,
    const Integer* const p,
    Real* const b,
    const Integer nw,
    Real* const work)
{
    Integer k, j;

    /* Buffer memory must be the same length as b. */
    if (n != nw)
    {
        return -1;
    }

    /* Rearrange right-hand-side vector b according to pivot vector p. Result stored in work. */
    for (k = 0; k != n; ++k)
    {
        work[k] = b[p[k]];
    }

    /* Forward substitution using L (lower triangular part of D). Result stored in b. */
    for (k = 0; k != n; ++k)
    {
        const Integer kn = k*n;
        Real sum = real(0);

        for (j = 0; j != k; ++j)
        {
            sum += D[kn + j] * b[j];
        }
        b[k] = work[k] - sum;
    }

    /* Back substitution using U. Overwrites b. */
    for (k = n - 1; k != -1; --k)
    {
        const Integer kn = k * n;
        Real sum = real(0);

        for (j = k + 1; j != n; ++j)
        {
            sum += D[kn + j] * b[j];
        }
        b[k] = (b[k] - sum) / D[kn + k]; /* Division safe as U is non-singular. */
    }

    return 0;
}

/*
* solve linear system
*
*/
ErrorSignal ESP_solveLinearEquations(Integer n, Real *A, Real *b, Integer *p, Real *w) {
    Boolean ok = true;

    ok = ((false != ok) && (false == ESP_hasNaN(b, n)))            ? true : false;
    ok = ((false != ok) && (false == ESP_hasNaN(A, n*n)))          ? true : false;
    ok = ((false != ok) && (false == LinFactorize(n, A, n, w, p))) ? true : false;
    ok = ((false != ok) && (false == LinSolve(n, A, p, b, n, w)))  ? true : false;

    if(true == ok) {
        return ErrorSignal_NONE;
    } else {
        Integer i;
        for(i=0; i<n; ++i) {
            b[i] = ESP_nan();
        }
        return ErrorSignal_SOLVE_LINEAR_EQUATIONS_FAILED;
    }
}

/**
 * find index i in monotonic increasing array x[] the index i between a (included)
 * and b (excluded), such that x[i]<=x_<x[i+1]
 */
static Integer binarySearch(Real x_, Real x[], Integer a, Integer b) {
    while((b-a) > 1)
    {
        Integer c = a - ((a-b)/2);
        if(x_<x[c]) {b = c;}
        else        {a = c;}
    }
    return a;
}

#if 0
reference: https://gitlab.ida.liu.se/emphysis/Specification/-/blob/master/0000_efmi_specification.adoc#user-content-language-specification-builtin-functions
#endif

Real interpolation1D(Real x1,
        Real x[], Integer nx1,
        Real y[], Integer ipo, Integer expo)
{
    Integer i = nx1 - 1;
    Real d;

    if((i <= 0)
    || (ipo  < 0) || (ipo  > 2)
    || (expo < 0) || (expo > 2))
    {
        return ESP_nan();
    } /* error */

    if (x1 <= x[0]) {
        if(expo == 1) {
            return y[0];
        } else {
            i = 0;
        }
    } else if(x1 >= x[i]) {
        if(expo == 1) {
            return y[i];
        } else {
            --i;
        }
    } else {
        i = binarySearch(x1, x, 0, i);
        if(ipo == 1) {return y[i];}
    }

    d = x[i+1] - x[i];
    if(isZero(d)) {return ESP_nan();} /* error */
    return y[i] + (((y[i+1] - y[i])*(x1 - x[i]))/d);
}

Real interpolation2D_(Real x1, Real x2,
        Real nx1_data[], Integer nx1,
        Real nx2_data[], Integer nx2,
        Real y_data[], Integer ipo, Integer expo)
{
    Integer i = nx1 - 1;
    Real a, b, c;

    if((i <= 0)
    || (ipo  < 0) || (ipo  > 2)
    || (expo < 0) || (expo > 2))
    {
        return ESP_nan();
    } /* error */

    if (x1 <= nx1_data[0]) {
        if(expo == 1) {
            return interpolation1D(x2,
                nx2_data, nx2,
                &y_data[0], ipo, expo);
        } else {
            i = 0;
        }
    } else if(x1 >= nx1_data[i]) {
        if(expo == 1) {
            return interpolation1D(x2,
                    nx2_data, nx2,
                    &y_data[i*nx2], ipo, expo);
        } else {
            --i;
        }
    } else {
        i = binarySearch(x1, nx1_data, 0, i);
        if(ipo == 1) {
            return interpolation1D(x2,
                    nx2_data, nx2,
                    &y_data[i*nx2], ipo, expo);
        }
    }

    c = nx1_data[i+1] - nx1_data[i];
    if(isZero(c))
    {
        return ESP_nan();
    } /* error */

    a = interpolation1D(x2,
            nx2_data, nx2,
            &y_data[(i+0)*nx2], ipo, expo);
    if(isNaN(a))
    {
        return ESP_nan();
    } /* error */

    b = interpolation1D(x2,
            nx2_data, nx2,
            &y_data[(i+1)*nx2], ipo, expo);
    if(isNaN(b))
    {
        return ESP_nan();
    } /* error */

    return a + (((b - a)*(x1 - nx1_data[i]))/c);
}

Real interpolation3D_(Real x1, Real x2, Real x3,
        Real *nx1_data, Integer nx1,
        Real *nx2_data, Integer nx2,
        Real *nx3_data, Integer nx3,
        Real *y_data,   Integer ipo, Integer expo) {
    Integer i = nx1 - 1;
    Real a, b, c;

    if((i <= 0)
    || (ipo  < 0) || (ipo  > 2)
    || (expo < 0) || (expo > 2))
    {
        return ESP_nan();
    } /* error */

    if (x1 <= nx1_data[0]) {
        if(expo == 1) {
            return interpolation2D_(x2, x3,
                    nx2_data, nx2,
                    nx3_data, nx3,
                    &y_data[0*nx2*nx3], ipo, expo);
        } else {
            i = 0;
        }
    } else if(x1 >= nx1_data[i]) {
        if(expo == 1) {
            return interpolation2D_(x2, x3,
                    nx2_data, nx2,
                    nx3_data, nx3,
                    &y_data[i*nx2*nx3], ipo, expo);
        } else {
            --i;
        }
    } else {
        i = binarySearch(x1, nx1_data, 0, i);
        if(ipo == 1) {
            return interpolation2D_(x2, x3,
                    nx2_data, nx2,
                    nx3_data, nx3,
                    &y_data[i*nx2*nx3], ipo, expo);
        }
    }

    c = nx1_data[i+1] - nx1_data[i];
    if(isZero(c))
    {
        return ESP_nan();
    } /* error */

    a = interpolation2D_(x2, x3,
            nx2_data, nx2,
            nx3_data, nx3,
            &y_data[(i+0)*nx2*nx3], ipo, expo);
    if(isNaN(a))
    {
        return ESP_nan();
    } /* error */

    b = interpolation2D_(x2, x3,
            nx2_data, nx2,
            nx3_data, nx3,
            &y_data[(i+1)*nx2*nx3], ipo, expo);
    if(isNaN(b))
    {
        return ESP_nan();
    } /* error */

    return a + (((b - a)*(x1 - nx1_data[i]))/c);
}

#ifdef __cplusplus
}
#endif
/* end of file */
