/*2021-03-01T17:48:00.470Z*/

/**********************************************************************************************************************
 * esp_alg.h
 *
 * generated by : embedded software producer/EfmiProductionCodeGen 1.0.0
 *
 * with types :
 *   Boolean   = unsigned int
 *   Integer   = signed int
 *   Real      = float
 *
 * with calling convention :
 *   Arguments = global var (efmiData)
 *   Ret. type = Boolean
 *
 * with names :
 *   Init Fcn  = efmiInitialize
 *   Step Fcn  = efmiDoStep
 *   Recal Fcn = efmiReInitialize
 *
 * with extra comments for :
 *   Code     = true
 *   Location = true
 *
 * with optimizations for :
 *   Indexing = true
 *   Parenth. = true (Misra: true)
 *   Id. len  = -1
 *   Tab size = 4
 *   Line len = 120
 **********************************************************************************************************************/

#ifndef ALG_H
#define ALG_H

#include <stdio.h>
#include <stdlib.h>

#include <math.h>
#include <float.h>
#include <limits.h>
#include <stdint.h>
#if defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 199901L) /* C99 */
#include <stdbool.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif

/****************************************************************************
 * Predefined signals
 ****************************************************************************/
typedef unsigned int                                ErrorSignal;
#define ErrorSignal_NONE                            (0u)
#define ErrorSignal_INVALID_ARGUMENT                (1u << 0)
#define ErrorSignal_OVERFLOW                        (1u << 1)
#define ErrorSignal_NAN                             (1u << 2)
#define ErrorSignal_SOLVE_LINEAR_EQUATIONS_FAILED   (1u << 3)
#define ErrorSignal_NO_SOLUTION_FOUND               (1u << 4)
#define ErrorSignal_UNSPECIFIED_ERROR               (1u << 5)

/****************************************************************************
 * Primitive types.
 ****************************************************************************/
typedef unsigned int    Boolean;
typedef signed int      Integer;
typedef float           Real;   /* 32 bits precision */

#define INT_SEL(i,l,ll) i
#define FLT_SEL(f,d,ld) f

#if !(defined(__cplusplus) || (defined(__STDC_VERSION__) && (__STDC_VERSION__ >= 199901L)))
#ifndef true
#define true                (1==1)
#endif

#ifndef false
#define false               (! true)
#endif
#endif

/****************************************************************************
 * Built-in functions
 ****************************************************************************/

/* allow automatic use of trunc/truncf/truncl & friends */
#define FLT_SFX(fcn)        FLT_SEL(fcn ## f, fcn /**/, fcn ## l)
#define INT_PFX(fcn)        INT_SEL(fcn /**/, l ## fcn, ll ## fcn)

/*****************************************************************************
 * Integer to Real and vice versa conversions
 ****************************************************************************/

/* converts x to boolean */
#ifndef boolean
#define boolean(x)          ((x)!=0)
#endif

/* Convert Integer i to Real. */
#ifndef real
#define real(i)             ((Real) (i))
#endif

/* Convert Real r to Integer by truncation (roundTowardsZero(r) - see appendix). */
#ifndef integer
#define integer(r)          ((Integer)roundTowardsZero(r))
#endif

#ifndef roundTowardsZero
#define roundTowardsZero(x) ((FLT_SFX(trunc))(x))
#endif

/* Round towards  minus infinity (also known as floor(r)). */
#ifndef roundDown
#define roundDown(x)        ((FLT_SFX(floor))(x))
#endif

/* Round towards plus infinity (also known as ceil(r)). */
#ifndef roundUp
#define roundUp(x)          ((FLT_SFX(ceil))(x))
#endif

/* roundHalfToEven (r)      Also known as convergent rounding, statistician's rounding, Dutch rounding. */
#ifndef roundHalfToEven
#define roundHalfToEven(x)  ((FLT_SFX(roundeven))(x))
#endif

/*
 Division of Integer variables i1, i2 with rounding to an integer
 divisionTowardsZero(i1,i2)
 integer(i1/i2) - same as div(i1,i2) in C89.
 */
#ifndef divisionTowardsZero
#define divisionTowardsZero(x,y)    (INT_PFX(div)((x),(y)).quot)
#endif

/*
 Integer remainder of division of Integer variables i1, i2
 remainderTowardsZero(i1,i2)
 Integer remainder of integer(i1/i2)- same as rem(i1,i2) in C89.
 */
#ifndef remainderTowardsZero
#define remainderTowardsZero(x,y)   (INT_PFX(div)((x),(y)).rem)
#endif

/*
 Remainder of division of Real variables r1, r2
 realRemainderTowardsZero(r1,r2)
 Real remainder of integer(r1/r2).
 */
#ifndef realRemainderTowardsZero
#define realRemainderTowardsZero(x) ((FLT_SFX(fmod))(x))
#endif

/*****************************************************************************
 * Relational functions of Real and Integer variables
 *****************************************************************************/

/* min(r1, r2)          Minimum of Real variables r1 and r2. */
#ifndef ESP_min
#define ESP_min(x,y)    ((FLT_SFX(fmin))((x),(y)))
#endif

/* max(r1, r2)          Maximum of Real variables r1 and r2. */
#ifndef ESP_max
#define ESP_max(x,y)    ((FLT_SFX(fmax))((x),(y)))
#endif

/* imin(i1,i2)          Minimum of Integer variables i1 and i2.*/
#ifndef imin
#define imin(a,b)       ((a)<(b) ? (a) : (b))
#endif
/* imax(i1,i2)          Maximum of Integer variables i1 and i2.*/
#ifndef imax
#define imax(a,b)       ((a)<(b) ? (b) : (a))
#endif

/****************************************************************************
 * Properties of the Integer variable type
 ****************************************************************************/

/* minInteger()         Target-specific smallest Integer. */
#ifndef minInteger
#define minInteger()    (INT_SEL(INT_MIN,LONG_MIN,LLONG_MIN))
#endif

/* maxInteger()         Target-specific largest Integer. */
#ifndef maxInteger
#define maxInteger()    (INT_SEL(INT_MAX,LONG_MAX,LLONG_MAX))
#endif

/****************************************************************************
 * Properties of the Real variable type
 ****************************************************************************/

/* minReal()            Target-specific smallest Real r <> minusInfinite(). */
#ifndef minReal
#define minReal()       (-maxReal())
#endif

/* maxReal()            Target-specific largest Real r <> plusInfinite(). */
#ifndef maxReal
#define maxReal()       (FLT_SEL(FLT_MAX,DBL_MAX,LDBL_MAX))
#endif

/* posMinReal()         Returns the smallest Real number r > 0. */
#ifndef posMinReal
#define posMinReal()    (FLT_SEL(FLT_MIN,DBL_MIN,LDBL_MIN))
#endif

/* epsReal()            Returns the largest Real number r > 0 such that 1.0+r == 1.0. */
#ifndef epsReal
#define epsReal()       (FLT_SEL(FLT_EPSILON,DBL_EPSILON,LDBL_EPSILON))
#endif

/* nan()                Target-specific quiet not-a-number representation (qNaN) */
extern Real ESP_nan(void);

/* isNaN(x)             = true, if x=NaN. */
#ifndef isNaN
#define isNaN(x)        (isnan(x))
#endif

/* minusInfinite()      Target-specific -oo representation.  */
#ifndef minusInfinite
#define minusInfinite() (-(plusInfinite()))
#endif

/* minusInfinite()      Target-specific +oo representation.  */
#ifndef plusInfinite
#define plusInfinite()  (real(INFINITY))
#endif

/* isInfinite(x)        = true, if x is infinite (= +infinity or -infinity)*/
#ifndef isInfinite
#define isInfinite(x)   (isinf(x))
#endif

/* isFinite(x)          = true, if x is finite (neither infinity nor NaN)*/
#ifndef isFinite
#define isFinite(x)     (isfinite(x))
#endif

/****************************************************************************
 * Multi-dimensional properties of Real
 ****************************************************************************/

/* hasNaN1D(x)          = true, if at least one element of vector x is NaN.*/
static inline Boolean ESP_hasNaN(Real *ptr, int n) {
    while((n != 0) && !isNaN(*ptr))
    {
        ++ptr;
        --n;
    }
    return n!=0 ? true : false;
}
#ifndef hasNaN1D
#define hasNaN1D(x)     ESP_hasNaN(&x[0], sizeof(x)/sizeof(Real))
#endif

/* hasNaN2D(x)          = true, if at least one element of matrix x is NaN.*/
#ifndef hasNaN2D
#define hasNaN2D(x)     ESP_hasNaN(&x[0][0], sizeof(x)/sizeof(Real))
#endif

/****************************************************************************
 * Basic mathematical functions of Real variables x
 ****************************************************************************/

/* euler() Target-specific, most-precise representation of e, Euler’s number. */
#ifndef euler
#define euler()         (real(M_E))
#endif

/* pi() Target-specific, most-precise representation of pi, the ratio of a circle’s
circumference to its diameter. */
#ifndef pi
#define pi()            (real(M_PI))
#endif

/* sign(x)              Sign of x (1.0, -1.0 or 0.0). */
#ifndef sign
#define sign(x)         ((x)<0 ? -1 : (x)>0 ? 1 : 0)
#endif

/* absolute(x)          Absolute value of Real variable x. */
#ifndef absolute
#define absolute(x)     ((FLT_SFX(fabs))(x))
#endif

/* fractional(x)        Fractional part of Real variable x. */
#ifndef fractional
#define fractional(x)   realRemainderTowardsZero(x,real(1))
#endif

/* safe_posdiv(xn, xd, eps) NaN-free: xn/(if xd>eps then xd else eps)*/
#ifndef safe_posdiv
#define safe_posdiv(xn,xd,eps)      ((xn)/ESP_max((xd),(eps)))
#endif

/* sqrt(x)              Square root of x (NaN if x < 0.0).*/
#ifndef ESP_sqrt
#define ESP_sqrt(x)     ((FLT_SFX(sqrt))(x))
#endif

/* safe_sqrt(x)         NaN-free: sqrt(if x>0 then x else 0)*/
#define safe_sqrt(x)    ESP_sqrt(ESP_max((x), 0))

/* exp(x)               Natural base exponential of x.*/
#ifndef ESP_exp
#define ESP_exp(x)      ((FLT_SFX(exp))(x))
#endif

/* log(x)               Natural logarithm of x (NaN if x < 0.0).*/
#ifndef ESP_log
#define ESP_log(x)      ((FLT_SFX(log))(x))
#endif

/* NaN-free: log(if x>0 then x else 0)*/
#define safe_log(x)     ESP_log(ESP_max((x),0))

/* log10(x)             Logarithm to base 10 of x (NaN if x < 0.0).*/
#ifndef ESP_log10
#define ESP_log10(x)    ((FLT_SFX(log10))(x))
#endif

/* safe_log10(x)        NaN-free: log10(if x>0 then x else 0)*/
#define safe_log10(x)   ESP_log10(ESP_max((x),0))

/* Trigonometric functions */

/* sin(x)               Sine of x.*/
#ifndef ESP_sin
#define ESP_sin(x)      ((FLT_SFX(sin))(x))
#endif

/* cos(x)               Cosine of x. */
#ifndef ESP_cos
#define ESP_cos(x)      ((FLT_SFX(cos))(x))
#endif

/* tan(x)               Tangent of x (NaN for x=pi/2, 3pi/2, ...). */
#ifndef ESP_tan
#define ESP_tan(x)      ((FLT_SFX(tan))(x))
#endif

/* safe_tan(x)          NaN-free: if x >= pi/2 then +inf elseif x <= -pi/2 then -inf else tan(x) */
#define clamp(x,mi,mx)  ESP_max(ESP_min((x),real(mx)),real(mi))
#define safe_tan(x)     ESP_tan(clamp((x),-M_PI_2,M_MI_2))

/* asin(x)              Inverse sine of x (NaN if x > 1.0 or x < -1.0). */
#ifndef ESP_asin
#define ESP_asin(x)     ((FLT_SFX(asin))(x))
#endif

/* safe_asin(x)         NaN-free: asin(if x > 1.0 then 1.0 elseif x < -1.0 then -1.0 else x) */
#define safe_asin(x)    ESP_asin(clamp((x),-1,1))

/* acos(x)              Inverse cosine of x (NaN if x > 1.0 or x < -1.0). */
#ifndef ESP_acos
#define ESP_acos(x)     ((FLT_SFX(acos))(x))
#endif

/* safe_acos(x)         NaN-free: acos(if x > 1.0 then 1.0 elseif x < -1.0 then -1.0 else x) */
#define safe_acos(x)    ESP_acos(clamp((x),-1,1))

/* atan(x)              Inverse tangent of x. Result is in the range -pi/2 <= y <= pi/2 */
#ifndef ESP_atan
#define ESP_atan(x)     ((FLT_SFX(atan))(x))
#endif

/* atan2(xn,xd)         Inverse tangent of xn/xd. Result is in the range -pi <= y <= pi (NaN if xn=0.0 and xd=0.0) */
#ifndef ESP_atan2
#define ESP_atan2(x,y)  ((FLT_SFX(atan2))(x,y))
#endif

/* sinh(x)              Hyperbolic sine. */
#ifndef ESP_sinh
#define ESP_sinh(x)     ((FLT_SFX(sinh))(x))
#endif

/* cosh(x)              Hyperbolic cosine. */
#ifndef ESP_cosh
#define ESP_cosh(x)     ((FLT_SFX(cosh))(x))
#endif

/* tanh(x)              Hyperbolic tangent. */
#ifndef ESP_tanh
#define ESP_tanh(x)     ((FLT_SFX(tanh))(x))
#endif

/*****************************************************************************
 * Systems of linear equations
 *****************************************************************************/
/*
 solveLinearEquations(A,b)
 Solve A*x=b for x
 (if no unique solution: all x[i] = NaN + error signal NO_UNIQUE_SOLUTION)

 (LU,pivots) = luFactorize(A)
 LU decomposition with partial pivoting of square matrix A.
 (if no unique solution: at least one LU[i,i] = 0 + error signal NO_UNIQUE_SOLUTION)

 x=luSolve(LU,pivots,b)
 Solve L*U*x =b[pivots] for x
 (if no unique solution: at least one x[i] = NaN + error signal NO_UNIQUE_SOLUTION)
 */
ErrorSignal ESP_solveLinearEquations(Integer n, Real *A, Real *b, Integer *p, Real *w);

/*****************************************************************************
 * Interpolation in 1D/2D/3D
 *****************************************************************************/

/*
 * y = interpolation1D(..)
 *
 * interpolation1D(x1,x1_data,nx1,y_data,ipo,expo)
 * Constant/linear interpolation in 1D with extrapolation
 */
Real interpolation1D(Real x1,
        Real x[], Integer nx1,
        Real y[], Integer ipo, Integer expo);

/*
 * y = interpolation2D(..)
 *
 * interpolation2D(x1,x2,x1_data,nx1,nx2_data,nx2,y_data,ipo,expo)
 * Constant/linear interpolation in 2D with extrapolation.
 */
Real interpolation2D_(Real x1, Real x2,
        Real nx1_data[], Integer nx1,
        Real nx2_data[], Integer nx2,
        Real y_data[], Integer ipo, Integer expo);

#define interpolation2D(x1, x2, nx1_data, nx1, nx2_data, nx2, y_data, ipo, expo) \
        interpolation2D_(x1, x2, nx1_data, nx1, nx2_data, nx2, &(y_data)[0][0], ipo, expo)

/*
 * y = interpolation3D(..)
 *
 * interpolation3D(x1,x2,x3,x1_data,nx1,nx2_data,nx2,nx3_data,nx3,y_data,ipo,expo)
 * Constant/linear interpolation in 3D with extrapolation.
*/

Real interpolation3D_(Real x1, Real x2, Real x3,
        Real nx1_data[], Integer nx1,
        Real nx2_data[], Integer nx2,
        Real nx3_data[], Integer nx3,
        Real y_data[], Integer ipo, Integer expo);

#define interpolation3D(x1, x2, nx1_data, nx1, nx2_data, nx2, nx3_data, nx3, y_data, ipo, expo) \
        interpolation3D_(x1, x2, nx1_data, nx1, nx2_data, nx2, &(y_data)[0][0][0], ipo, expo)

/*****************************************************************************
 * extra stuff
 *****************************************************************************/

#ifdef __TINYC__
/* work around tcc bug in which trunc() fails to work after few calls */
#undef roundTowardsZero
static inline Real roundTowardsZero(Real x)
{
    return (x >= 0) ? roundDown(x) : roundUp(x);
}
#endif

static inline Real ESP_square(Real x)
{
    return x*x;
}

static inline Real ESP_cube(Real x)
{
    return x*x*x;
}

static inline Real ESP_ipow(Real x, Integer y)
{
    Real z = real(1);

    if(y < 0)
	{
		y = -y;
		x = z/x;
	}

    while(y>0)
    {
        if((y & 1U)!=0)
        {
        	z *= x;
        }
        x *= x;
        y /= 2;
    }

    return z;
}

#ifndef ESP_pow
#define ESP_pow(x,y)                        ((FLT_SFX(pow))(x,y))
#endif

#ifndef ESP_div
#define ESP_div(x,y)                        (divisionTowardsZero((x),(y)))
#endif

#ifndef ESP_mod
#define ESP_mod(x,y)                        ((x) -  ((y)*roundDown((x)/(y))))
#endif

#ifndef realRemainderDown
#define realRemainderDown(x,y)              ((x) -  ((y)*roundDown((x)/(y))))
#endif

#ifndef realRemainderUp
#define realRemainderUp(x,y)                ((x) -  ((y)*roundUp((x)/(y))))
#endif

#ifndef realRemainderAwayZero
#define realRemainderAwayZero(x,y)          ((x) -  ((y)*roundAwayZero((x)/(y))))
#endif

#ifndef realRemainderHalfDown
#define realRemainderHalfDown(x,y)          ((x) -  ((y)*roundAwayHalfDown((x)/(y))))
#endif

#ifndef realRemainderHalfUp
#define realRemainderHalfUp(x,y)            ((x) -  ((y)*roundAwayHalfUp((x)/(y))))
#endif

#ifndef realRemainderHalfTowardsZero
#define realRemainderHalfTowardsZero(x,y)   ((x) -  ((y)*roundHalfTowardsZero((x)/(y))))
#endif

#ifndef realRemainderHalfAwayZero
#define realRemainderHalfAwayZero(x,y)      ((x) -  ((y)*roundHalfAwayZero((x)/(y))))
#endif

#ifndef realRemainderHalfToEven
#define realRemainderHalfToEven(x,y)        ((x) -  ((y)*roundHalfToEven((x)/(y))))
#endif

#ifndef realRemainderHalfToOdd
#define realRemainderHalfToOdd(x,y)         ((x) -  ((y)*roundHalfToOdd((x)/(y))))
#endif

#ifndef ESP_HALF
#define ESP_HALF                real(0.5)
#endif

#ifndef roundAwayZero
#define roundAwayZero(x)        (((x)<=real(0)) ? roundDown(x) : roundUp(x))
#endif

#ifndef roundHalfDown
#define roundHalfDown(x)        roundUp((x) - ESP_HALF)
#endif

#ifndef roundHalfUp
#define roundHalfUp(x)          roundDown((x) + ESP_HALF)
#endif

#ifndef roundHalfTowardsZero
#define roundHalfTowardsZero(x) roundAwayZero((x) - (sign(x)*ESP_HALF))
#endif

#ifndef roundHalfAwayZero
#define roundHalfAwayZero(x)    roundTowardsZero((x) + (sign(x)*ESP_HALF))
#endif

#ifndef roundHalfToOdd
#define roundHalfToOdd(r)       ((roundHalfDown(r) < roundHalfUp(r)) \
        ? ( (integer(remainder((r) + ESP_HALF, real(2))) == 0)       \
                ? (r) - ESP_HALF                                     \
                : (r) + ESP_HALF )                                   \
        : roundHalfDown(r))
#endif

#ifndef divisionEuclidean
#define divisionEuclidean(dividend, divisor)    \
        integer(((divisor) > 0 ? roundDown(real(dividend) / real(divisor)) : \
        roundUp(real(dividend) / real(divisor))))
#endif

#ifdef __cplusplus
}
#endif

#endif /* ALG_H */
/* end of file */
